/* tcpsend:
    This is a diagnostic tool for use by a network administrator.

    Copyright (c) 2002, The Trustees of Princeton University, All Rights Reserved.
*/



#ifdef HAVE_CONFIG_H
# include "config.h"
#endif

#include "defs.h"

#include "tcpsend.h"
#include "utils.h"

#ifndef lint
static const char rcsid[] = "tcpsend version " VERSION;
static const char copyright[] = "Copyright 2003, The Trustees of Princeton University.  All rights reserved.";
static const char contact[] = "networking@princeton.edu";
#endif

/* options, args, etc. */
char *prog;
int debug = 0;
int suppress_warnings = 0;
int quiet = 0;
int count = 1;
unsigned int pause_time = 1;

char *interface_name = NULL;
struct ether_addr ether_source_address;
struct ether_addr ether_dest_address;
struct in_addr ip_src_address;
struct in_addr ip_dest_address;
unsigned char ip_ttl = (unsigned char) IP_TTL_DEFAULT;
unsigned short ip_id = (unsigned short) IP_ID_DEFAULT;
unsigned short ip_frag = 0;
int ip_flags = 0;
unsigned short tcp_src_port = 0;
unsigned short tcp_dest_port = 0;
unsigned long tcp_seq_num = 0;
unsigned long tcp_ack_num = 0;
unsigned short tcp_urg_ptr = 0;
unsigned short tcp_win = (unsigned short) TCP_WIN_DEFAULT;
int tcp_flags = 0;

int ether_source_address_set = 0;
int ether_dest_address_set = 0;
int ip_src_address_set = 0;
int ip_dest_address_set = 0;
int ip_ttl_set = 0;
int ip_id_set = 0;
int ip_frag_set = 0;
int tcp_src_port_set = 0;
int tcp_dest_port_set = 0;
int tcp_seq_num_set = 0;
int tcp_ack_num_set = 0;
int tcp_urg_ptr_set = 0;
int tcp_win_set = 0;



int 
main(int argc, char **argv)
{
	/* for getopt() option and arg parsing */
	int c, errflag = 0;
	extern char *optarg;
	extern int optind, opterr, optopt;

	int do_send_packet;

	/* get progname = last component of argv[0] */
	prog = strrchr(argv[0], '/');
	if (prog)
		prog++;
	else 
		prog = argv[0];

	while ((c = getopt(argc, argv, "A:B:c:D:d:E:e:f:hi:p:qS:s:T:U:vW:wY:y:01234567")) != EOF) {
		switch (c) {
			
			case 'A': {
				/* XXX lacks error checking */
				tcp_ack_num = (unsigned long) atol(optarg);
				tcp_ack_num_set++;
				break;
			}

			case 'B': {
				/* XXX lacks error checking */
				tcp_seq_num = (unsigned long) atol(optarg);
				tcp_seq_num_set++;
				break;
			}

			case 'c': {
				char *stmp = optarg;
				if ((sscanf(stmp, "%d", &count) != 1) || (count < 1)) {
					fprintf(stderr, "%s: invalid packet count '%s', must be integer > 0\n", prog, optarg);
					count = 1;
					errflag++;
				}
				break;
			}

			case 'D':
				/* XXX lacks error checking */
				ip_id = (unsigned short) atoi(optarg);
				ip_id_set++;
				break;

			case 'd': {
				char *stmp = optarg;
				if ((sscanf(stmp, "%d", &debug) != 1) || (debug < 0)) {
					fprintf(stderr, "%s: invalid debug level '%s'\n", prog, optarg);
					debug = 0;
					errflag++;
				}
				break;
			}

			case 'E': {
				struct ether_addr *eaddr_tmp;
				if ((eaddr_tmp = ether_aton(optarg)) == NULL) {
					fprintf(stderr, "%s: invalid ether_source_address '%s'\n", prog, optarg);
					errflag++;
				} else { 
					bcopy(eaddr_tmp, &ether_source_address, sizeof(ether_source_address));
					ether_source_address_set++;
				}
				break;
			}

			case 'e': {
				struct ether_addr *eaddr_tmp;
				if ((eaddr_tmp = ether_aton(optarg)) == NULL) {
					fprintf(stderr, "%s: invalid ether_dest_address '%s'\n", prog, optarg);
					errflag++;
				} else { 
					bcopy(eaddr_tmp, &ether_dest_address, sizeof(ether_dest_address));
					ether_dest_address_set++;
				}
				break;
			}

			case 'f':
				/* XXX lacks error checking */
				ip_frag = (unsigned short) atoi(optarg);
				ip_frag &= IP_OFFMASK;
				ip_frag_set++;
				break;

			case 'h':
				usage();
				exit(0);

			case 'i':
				interface_name = optarg;
				break;

			case 'p': {
				char *stmp = optarg;
				if ((sscanf(stmp, "%u", &pause_time) != 1) || (pause_time < 1)) {
					fprintf(stderr, "%s: invalid pause_time '%s', must be integer > 0\n", prog, optarg);
					pause_time = 1;
					errflag++;
				}
				break;
			}

			case 'q':
				quiet++;
				break;

			case 'S':
				if (inet_aton(optarg, &ip_src_address) == 0) {
					fprintf(stderr, "%s: invalid ip_src_address '%s'\n", prog, optarg);
					errflag++;
				} else {
					ip_src_address_set++;
				}
				break;

			case 's':
				if (inet_aton(optarg, &ip_dest_address) == 0) {
					fprintf(stderr, "%s: invalid ip_dest_address '%s'\n", prog, optarg);
					errflag++;
				} else {
					ip_dest_address_set++;
				}
				break;

			case 'T':
				/* XXX lacks error checking */
				ip_ttl = (unsigned char) atoi(optarg);
				ip_ttl_set++;
				break;

			case 'U':
				/* XXX lacks error checking */
				tcp_urg_ptr = (unsigned short) atoi(optarg);
				tcp_urg_ptr_set++;
				break;


			case 'v':
				printf("tcpsend version %s\n", VERSION);
				exit(0);

			case 'W':
				/* XXX lacks error checking */
				tcp_win = (unsigned short) atoi(optarg);
				tcp_win_set++;
				break;

			case 'w':
				suppress_warnings++;
				break;

			case 'Y':
				/* XXX lacks error checking */
				tcp_src_port = (unsigned short) atoi(optarg);
				tcp_src_port_set++;
				break;

			case 'y':
				/* XXX lacks error checking */
				tcp_dest_port = (unsigned short) atoi(optarg);
				tcp_dest_port_set++;
				break;

			case '0':
				tcp_flags |= TH_SYN;
				break;

			case '1':
				tcp_flags |= TH_ACK;
				break;

			case '2':
				tcp_flags |= TH_FIN;
				break;

			case '3':
				tcp_flags |= TH_RST;
				break;

			case '4':
				tcp_flags |= TH_URG;
				break;

			case '5':
				tcp_flags |= TH_PUSH;
				break;

			case '6':
				ip_flags |= IP_DF;
				break;

			case '7':
				ip_flags |= IP_MF;
				break;

			case '?':
				usage();
				exit(0);

			default:
				errflag++;
				break;

		} /* switch (c) */
	} /* while c=getopt() */

	/* a valid ether_dest_address option is actually required */
	if (!ether_dest_address_set) {
		fprintf(stderr, "%s: ether_dest_address is required\n", prog);
		errflag++;
	}

	/* a valid ip_dest_address option is actually required */
	if (!ip_dest_address_set) {
		fprintf(stderr, "%s: ip_dest_address is required\n", prog);
		errflag++;
	}

	if (errflag) {
		if (!quiet) {
			fprintf(stderr, "\n"); /* separate error message we've already printed from usage message */
			usage();
		}
		exit(1);
	}
 
	/* we're done parsing commandline opts and args */

	/* fill in settings, taking into account commandline opts and args */
	fillin_settings();
	
	/* if we're allowed to warn about dangerous settings (warnings aren't suppressed)
	   and the settings are actually dangerous,
	   then prompt to see if we should still send the packet
	*/
	do_send_packet = 1; /* default to send */
	if (!suppress_warnings && is_dangerous_settings())
		do_send_packet = send_packet_prompt();


	if (do_send_packet) {
		while (count--) {
			if (send_packet()) {
				/* the send succeeded */
				if (!quiet)
					fprintf(stdout, "packet sent\n");
			}
			if (count) /* there are more packets to send */
				sleep(pause_time);
		}
	} else
		fprintf(stdout, "send cancelled\n");

	cleanup();
	exit(0);
}


/* prompt if we should send the packet (default 'no').
   Return 1 for yes, 0 for no.
*/
int
send_packet_prompt(void)
{
	char buf[MAXLINE], response[MAXLINE];

	/* preceed prompt with newline since we must have printed some warning(s) earlier */
	fprintf(stdout, "\nSend anyway? [no] ");

	if (fgets(buf, sizeof(buf), stdin) == NULL) {
		/* newline, EOF, or error */
		return 0; /* cancel sending packet */
	}
	response[0] = '\0';
	if (sscanf(buf, "%s", response) < 1) {
		/* newline, EOF, or error */
		return 0; /* cancel sending packet */
	}
	if ((response[0] == 'y') || (response[0] == 'Y'))
		return 1; /* send packet */
	else
		return 0; /* cancel sending packet */
}


/* cleanup tasks at exit */
void
cleanup(void)
{

	if (nd) {
		if (libnet_close_link_interface(nd) < 0) 
			fprintf(stderr, "%s: error closing interface '%s'\n", prog, interface_name);
	}
	return;
}


/* print usage message and return */
void
usage(void)
{

	char *usage= "Usage: ";
	char *usage_and_progname;
	int usage_and_progname_len;

	/* build beginning of the usage message */
	usage_and_progname_len = strlen(usage) + strlen(prog) +1; /* 1 = trailing space */
	usage_and_progname = (char *) smalloc(usage_and_progname_len +1, 0); /* 1 = trailing NUL */
	snprintf(usage_and_progname, usage_and_progname_len+1, "%s%s ", usage, prog);

	fprintf(stderr, "%s", usage_and_progname);
	fprintf(stderr, "[-c count] [-d debuglevel] [-h] [-p pause_time] [-q] [-v] [-w]\n");
	fprintf(stderr, "%*s[-i interface_name]\n", usage_and_progname_len, "");
	fprintf(stderr, "%*s[-E ether_source_address]  -e ether_dest_address\n", usage_and_progname_len, "");
	fprintf(stderr, "%*s[-S ip_src_address]        -s ip_dest_address\n", usage_and_progname_len, "");
	fprintf(stderr, "%*s[-T ip_ttl] [-D ip_id] [-f ip_frag] [-67]\n", usage_and_progname_len, "");
	fprintf(stderr, "%*s[-Y tcp_src_port]         [-y tcp_dest_port]\n", usage_and_progname_len, "");
	fprintf(stderr, "%*s[-B tcp_seq_num] [-A tcp_ack_num] [-W tcp_win] [-U tcp_urg_ptr] [-012345]\n", usage_and_progname_len, "");

	free(usage_and_progname);

	fprintf(stderr, "where:\n");
	fprintf(stderr, "   -c count                       count of packets to send (defaults to 1)\n");
	fprintf(stderr, "   -d debuglevel                  enable debugging at specified level (not presently used)\n");
	fprintf(stderr, "   -h                             display this help message then exit\n");
	fprintf(stderr, "   -p pause_time                  seconds to pause between sending each packet when count > 1 (defaults to 1)\n");
	fprintf(stderr, "   -q                             quiet output\n");
	fprintf(stderr, "   -v                             display version number then exit\n");
	fprintf(stderr, "   -w                             suppress warnings/prompt regarding dangerous settings\n");
	fprintf(stderr, "   -i interface_name              specify outgoing interface\n");

	fprintf(stderr, "  Ethernet Header:\n");
	fprintf(stderr, "   -E ether_source_address        override Ethernet frame source address (defaults to outgoing interface)\n");
	fprintf(stderr, "   -e ether_dest_address          specify Ethernet frame dest address (required)\n");

	fprintf(stderr, "  IP Header:\n");
	fprintf(stderr, "   -S ip_src_address              override IP source address (defaults to outgoing interface)\n");
	fprintf(stderr, "   -s ip_dest_address             specify IP dest address (required)\n");
	fprintf(stderr, "   -T ip_ttl                      override IP ttl (defaults to %d)\n", IP_TTL_DEFAULT);
	fprintf(stderr, "   -D ip_id                       override IP id (defaults to %d)\n", IP_ID_DEFAULT);
	fprintf(stderr, "   -f ip_frag                     override IP frag (defaults to 0)\n");
	fprintf(stderr, "   -6                             set IP DF flag\n");
	fprintf(stderr, "   -7                             set IP MF flag\n");

	fprintf(stderr, "  TCP Header:\n");
	fprintf(stderr, "   -Y tcp_src_port                override TCP src port (defaults to 0)\n");
	fprintf(stderr, "   -y tcp_dest_port               override TCP dest port (defaults to 0)\n");
	fprintf(stderr, "   -B tcp_seq_num                 override TCP seq number (defaults to 0)\n");
	fprintf(stderr, "   -A tcp_ack_num                 override TCP ack number (defaults to 0)\n");
	fprintf(stderr, "   -U tcp_urg_ptr                 override TCP urg pointer (defaults to 0)\n");
	fprintf(stderr, "   -W tcp_win                     override TCP window size (defaults to %d)\n", TCP_WIN_DEFAULT);
	fprintf(stderr, "   -0                             set TCP SYN flag\n");
	fprintf(stderr, "   -1                             set TCP ACK flag\n");
	fprintf(stderr, "   -2                             set TCP FIN flag\n");
	fprintf(stderr, "   -3                             set TCP RST flag\n");
	fprintf(stderr, "   -4                             set TCP URG flag\n");
	fprintf(stderr, "   -5                             set TCP PSH flag\n");

	return;
}
