/*
 * fillin_settings()
 * is_dangerous_settings()
 * send_packet()
 * Copyright (c) 2001-2002, The Trustees of Princeton University, All Rights Reserved.
 */

#ifdef HAVE_CONFIG_H
# include "config.h"
#endif

#include "defs.h"

#include "tcpsend.h"

/* for libnet */
char libnet_errbuf[LIBNET_ERRBUF_SIZE];
struct libnet_link_int *nd = NULL;


/* Fill in settings, taking into account args and options.
   Also opens nd, the link interface.
*/
void 
fillin_settings(void)
{

	if (interface_name == NULL) {
		/* ask libnet to select the first interface that is up (other than loopback) */
		struct sockaddr_in sin;
		if (libnet_select_device(&sin, &interface_name, libnet_errbuf) == -1) {
			fprintf(stderr, "%s: error selecting an interface: %s\n", prog, libnet_errbuf);
			cleanup();
			exit(1);
		}
	}

	if ((nd = libnet_open_link_interface(interface_name, libnet_errbuf)) == NULL) {
		fprintf(stderr, "%s: error opening interface '%s': %s\n", prog, interface_name, libnet_errbuf);
		cleanup();
		exit(1);
	}

	if (!ip_src_address_set) {
		/* ip_src_address defaults to that of outgoing interface */
		/* Note libnet_get_ipaddr() returns host-byte order.  Versions of libnet older than 1.0.2
		   incorrectly document it as returning network-byte order. */
		if ((ip_src_address.s_addr = htonl(libnet_get_ipaddr(nd, interface_name, libnet_errbuf))) == 0) {
			fprintf(stderr, "%s: can't determine IP address of interface '%s': %s\n", prog, interface_name, libnet_errbuf);
			cleanup();
			exit(1);
		}
	}

	if (!ether_source_address_set) { 
		struct ether_addr *my_eaddr;

		/* set my_eaddr to that of the outgoing interface */
		if ((my_eaddr =libnet_get_hwaddr(nd, interface_name, libnet_errbuf)) == NULL) {
			fprintf(stderr, "%s: can't determine hardware address of interface '%s': %s\n", prog, interface_name, libnet_errbuf);
			cleanup();
			exit(1);
		}

		bcopy(my_eaddr, &ether_source_address, sizeof(ether_source_address));
	}

	
	if (!quiet) {
		fprintf(stdout, "Using settings:\n");
		fprintf(stdout, "  interface %s\n", interface_name);
		fprintf(stdout, "  ether_source_address %s\n", ether_ntoa(&ether_source_address));
		fprintf(stdout, "  ether_dest_address %s\n", ether_ntoa(&ether_dest_address));
		fprintf(stdout, "  ip_src_address %s\n", inet_ntoa(ip_src_address));
		fprintf(stdout, "  ip_dest_address %s\n", inet_ntoa(ip_dest_address));
		fprintf(stdout, "  ip_ttl %u\n", (unsigned int) ip_ttl);
		fprintf(stdout, "  ip_id %u\n", ip_id);
		fprintf(stdout, "  ip_frag %u\n", ip_frag);
		fprintf(stdout, "  ip_flags: %s%s\n", 
				(ip_flags & IP_DF) ? "dontfrag " : "",
				(ip_flags & IP_MF) ? "morefrags " : ""
			);
		fprintf(stdout, "  tcp_src_port %u\n", tcp_src_port);
		fprintf(stdout, "  tcp_dest_port %u\n", tcp_dest_port);
		fprintf(stdout, "  tcp_seq_num %lu\n", tcp_seq_num);
		fprintf(stdout, "  tcp_ack_num %lu\n", tcp_ack_num);
		fprintf(stdout, "  tcp_urg_ptr %u\n", tcp_urg_ptr);
		fprintf(stdout, "  tcp_win %u\n", tcp_win);
		fprintf(stdout, "  tcp_flags: %s%s%s%s%s%s\n", 
				(tcp_flags & TH_SYN) ? "S" : "",
				(tcp_flags & TH_ACK) ? "A" : "",
				(tcp_flags & TH_FIN) ? "F" : "",
				(tcp_flags & TH_RST) ? "R" : "",
				(tcp_flags & TH_URG) ? "U" : "",
				(tcp_flags & TH_PUSH) ? "P" : ""
			);
	}
	
	return;
}


/* Check for dangerous settings; return 1 if any found, else return 0.
   Print a warning for each dangerous item found.
 */
int
is_dangerous_settings(void)
{
	int danger_found = 0;


	/* An ether_src should never be multicast or broadcast.
	   Imagine what happens if bridges/switches learn ff:ff:ff:ff:ff:ff is located on a specific port!
	*/
	if (ether_source_address.ether_addr_octet[0] & 0x01) {
		if (danger_found) /* this is not our first warning */
			fprintf(stdout, "\n");
		fprintf(stdout, "*** Warning: ether_source_address '%s' is broadcast or multicast\n(not a valid Ethernet source).\n", 
			ether_ntoa(&ether_source_address));
		fprintf(stdout, "A careless switch/bridge might mistakenly insert this into its\nforwarding table.\n");
		danger_found++;
	}


	/* The ether_dest_address should not be multicast or broadcast, since a TCP message
	   should only be sent to a specific host.  
	   If it is one of these, let's provide some warning.
	*/
	if (ether_dest_address.ether_addr_octet[0] & 0x01) {
		if (danger_found) /* this is not our first warning */
			fprintf(stdout, "\n");
		fprintf(stdout, "*** Warning: ether_dest_address '%s' is broadcast or multicast\n.\n", 
			ether_ntoa(&ether_dest_address));
		fprintf(stdout, "TCP should only be sent to a unicast destination.\n");
		danger_found++;
	}


	/* IPdst should be unicast, since an TCP message
	   should only be sent to a specific host.
	   If it is one of these, let's provide some warning.
	*/
	if (IN_CLASSD(ip_dest_address.s_addr)) {
		if (danger_found) /* this is not our first warning */
			fprintf(stdout, "\n");
		fprintf(stdout, "*** Warning: ip_dest_address '%s' is an IP multicast address.\n", inet_ntoa(ip_dest_address));
		fprintf(stdout, "TCP should only be sent to a unicast destination.\n");
		danger_found++;
	}
	if (IN_EXPERIMENTAL(ip_dest_address.s_addr)) {
		if (danger_found) /* this is not our first warning */
			fprintf(stdout, "\n");
		fprintf(stdout, "*** Warning: ip_dest_address '%s' is an IP experimental (Class E) address.\n", inet_ntoa(ip_dest_address));
		fprintf(stdout, "TCP should only be sent to a unicast destination.\n");
		danger_found++;
	}
	if (ip_dest_address.s_addr == INADDR_BROADCAST) {
		if (danger_found) /* this is not our first warning */
			fprintf(stdout, "\n");
		fprintf(stdout, "*** Warning: ip_dest_address '%s' is the IP limited\nbroadcast address.\n", inet_ntoa(ip_dest_address));
		fprintf(stdout, "TCP should only be sent to a unicast destination.\n");
		danger_found++;
	}
	if (ip_dest_address.s_addr == INADDR_ANY) {
		if (danger_found) /* this is not our first warning */
			fprintf(stdout, "\n");
		fprintf(stdout, "*** Warning: ip_dest_address '%s' is the IP unspecified address.\n", inet_ntoa(ip_dest_address));
		fprintf(stdout, "TCP should only be sent to a unicast destination.\n");
		danger_found++;
	}
	if (MY_IN_LOOPBACK(ip_dest_address.s_addr)) {
		if (danger_found) /* this is not our first warning */
			fprintf(stdout, "\n");
		fprintf(stdout, "*** Warning: ip_dest_address '%s' is on the loopback network.\n", inet_ntoa(ip_dest_address));
		fprintf(stdout, "No IP packet should appear on the network with a loopback destination.\n");
		danger_found++;
	}
	/* XXX we should also check to see if ip_dest_address is any of the other flavors of IP broadcast
	   addresses valid on this network.  
	*/


	/* IPsrc should be unicast.
	*/
	if (IN_CLASSD(ip_src_address.s_addr)) {
		if (danger_found) /* this is not our first warning */
			fprintf(stdout, "\n");
		fprintf(stdout, "*** Warning: ip_src_address '%s' is an IP multicast address.\n", inet_ntoa(ip_src_address));
		fprintf(stdout, "TCP should only be sent from a unicast source.\n");
		danger_found++;
	}
	if (IN_EXPERIMENTAL(ip_src_address.s_addr)) {
		if (danger_found) /* this is not our first warning */
			fprintf(stdout, "\n");
		fprintf(stdout, "*** Warning: ip_src_address '%s' is an IP experimental (Class E) address.\n", inet_ntoa(ip_src_address));
		fprintf(stdout, "TCP should only be sent from a unicast source.\n");
		danger_found++;
	}
	if (ip_src_address.s_addr == INADDR_BROADCAST) {
		if (danger_found) /* this is not our first warning */
			fprintf(stdout, "\n");
		fprintf(stdout, "*** Warning: ip_src_address '%s' is the IP limited\nbroadcast address.\n", inet_ntoa(ip_src_address));
		fprintf(stdout, "TCP should only be sent from a unicast source.\n");
		danger_found++;
	}
	if (ip_src_address.s_addr == INADDR_ANY) {
		if (danger_found) /* this is not our first warning */
			fprintf(stdout, "\n");
		fprintf(stdout, "*** Warning: ip_src_address '%s' is the IP unspecified address.\n", inet_ntoa(ip_src_address));
		fprintf(stdout, "TCP should only be sent from a unicast source.\n");
		danger_found++;
	}
	if (MY_IN_LOOPBACK(ip_src_address.s_addr)) {
		if (danger_found) /* this is not our first warning */
			fprintf(stdout, "\n");
		fprintf(stdout, "*** Warning: ip_src_address '%s' is on the loopback network.\n", inet_ntoa(ip_src_address));
		fprintf(stdout, "No IP packet should appear on the network with a loopback source.\n");
		danger_found++;
	}
	/* XXX we should also check to see if ip_src_address is any of the other flavors of IP broadcast
	   addresses valid on this network.  
	*/

	return danger_found;
}


/* construct and send packet.
   Return 0 on success, 1 on failure.
*/
int
send_packet(void)
{
	u_char *write_packet = NULL;								/* packet buffer for writing */
	int ether_tcp_packet_size = LIBNET_ETH_H + LIBNET_IP_H + LIBNET_TCP_H;	/* size of packet we will write */
	int bytes_written;

	/* allocate packet buffer space and zero it */
	if (libnet_init_packet(ether_tcp_packet_size, &write_packet) == -1) {
		fprintf(stderr, "%s: libnet_init_packet failed\n", prog);
		if (write_packet != NULL) /* just in case something got allocated */
			libnet_destroy_packet(&write_packet);
		return(0); /* failure */
	}

	/* build the Ethernet header */
	if (libnet_build_ethernet(
		ether_dest_address.ether_addr_octet,		/* ether_dst */
		ether_source_address.ether_addr_octet,		/* ether_src */
		ETHERTYPE_IP,								/* ethertype */
		NULL, 0,									/* no optional payload */
		write_packet								/* start of ethernet header in packet */
	) < 0) {
		fprintf(stderr, "%s: libnet_build_ethernet failed\n", prog);
		libnet_destroy_packet(&write_packet);
		return(0); /* failure */
	}

	/* build the IP header and payload */
	if (libnet_build_ip(
		LIBNET_TCP_H,								/* IP packet len (not including IP header) */
		0,											/* tos */
		ip_id,										/* id */
		ip_flags | ip_frag ,						/* flags and frag_offset bits */
		ip_ttl,										/* ttl */
		IPPROTO_TCP,								/* protocol */
		ip_src_address.s_addr,						/* src address */
		ip_dest_address.s_addr,						/* dst address */
		NULL, 0,									/* no optional payload */
		write_packet + LIBNET_ETH_H					/* start of ip header in packet */
	) < 0) {
		fprintf(stderr, "%s: libnet_build_ip failed\n", prog);
		libnet_destroy_packet(&write_packet);
		return(0); /* failure */
	}

	/* build the TCP header and payload */
	if (libnet_build_tcp(
			tcp_src_port,							/* source port */
			tcp_dest_port,							/* dest port */
			tcp_seq_num,							/* sequence number */
			tcp_ack_num,							/* ack number */
			(unsigned char) tcp_flags,				/* flags */
			tcp_win,								/* window size */
			tcp_urg_ptr,							/* urgent pointer */
			NULL, 0,								/* no optional payload */
			write_packet + LIBNET_ETH_H + LIBNET_IP_H	/* start of TCP header in packet */
	) < 0) {
		fprintf(stderr, "%s: libnet_build_tcp failed\n", prog);
		libnet_destroy_packet(&write_packet);
		return(0); /* failure */
	}


	/* compute and insert the IP checksum */
	if (libnet_do_checksum(
		write_packet + LIBNET_ETH_H,				/* start of IP header */
		IPPROTO_IP,									/* compute and insert checksum for this transport protocol */
		LIBNET_IP_H									/* IP header length */
	) < 0) {
		fprintf(stderr, "%s: libnet_do_checksum(IP header) failed\n", prog);
		libnet_destroy_packet(&write_packet);
		return(0); /* failure */
	}

	/* compute and insert the TCP header checksum */
	if (libnet_do_checksum(
			write_packet + LIBNET_ETH_H,			/* start of TCP header */
			IPPROTO_TCP,							/* compute and insert checksum for this transport protocol */
			LIBNET_TCP_H							/* IP packet length (not including IP header */
	) < 0) {
		fprintf(stderr, "%s: libnet_do_checksum(TCP header) failed\n", prog);
		libnet_destroy_packet(&write_packet);
		return(0); /* failure */
	}

	/* write the packet */
	bytes_written =  libnet_write_link_layer(nd, interface_name, write_packet, ether_tcp_packet_size);
	if (bytes_written < ether_tcp_packet_size)
		fprintf(stderr, "%s: libnet_write_link_layer: bytes written: %d (expected %d)\n", prog, bytes_written, ether_tcp_packet_size);

	libnet_destroy_packet(&write_packet);

	return(1); /* success */
}
